"use strict";
const log = require("fary-log");
const koaHttpProxy = require("koa-better-http-proxy");
const ENV = require("../utils/env");
const isDomain = require("../utils/is-domain");
const serverConfig = require("../utils/server-config");
const { IS_DEV, IS_SERVER_PRODUCTION } = ENV;
const serverConfigPath = serverConfig.path;
const serverConfigValue = serverConfig.value;
const logger = log(('fary:proxy'));
/**
 * load proxy config
 * @return {*} proxy result
 */
function getProxyConfig() {
    if (!IS_DEV) {
        return serverConfigValue.proxy;
    }
    // in development mode, clear the cache every time when invoked to improve developing experience!
    delete require.cache[serverConfigPath];
    const newServerConfig = require(serverConfigPath);
    logger.log('Got newest proxy mapping:', newServerConfig.proxy);
    return newServerConfig.proxy;
}
module.exports = () => {
    return async function ProxyMiddleware(ctx, next) {
        const url = ctx.url;
        let proxyTarget = '';
        let proxyConfig = getProxyConfig();
        // only when non *real production environment*, allow the proxy debugging
        if (!IS_SERVER_PRODUCTION) {
            // if the `proxyconfig` header is present, then merge it to `proxyConfig` for debugging
            const headerProxyConfigStr = ctx.headers.proxyconfig;
            if (headerProxyConfigStr) {
                let headerProxyConfig;
                logger.log(`'proxyconfig' detected in headers: '${headerProxyConfigStr}'`);
                try {
                    headerProxyConfig = JSON.parse(headerProxyConfigStr);
                }
                catch (e) {
                    logger.warn('proxyconfig parsing FAILED, please check it.');
                    headerProxyConfig = {};
                }
                proxyConfig = Object.assign({}, proxyConfig, headerProxyConfig);
                logger.log('Merged proxy config is:', proxyConfig);
            }
        }
        // find the one which match the prefix of `url`
        for (const prefix of Object.keys(proxyConfig)) {
            if (url.startsWith(prefix)) {
                const target = proxyConfig[prefix];
                // avoid the multiple proxy problem when local development
                if (!IS_DEV || !isDomain(target)) {
                    ctx.url = url.replace(prefix, '');
                }
                proxyTarget = target;
                logger.log(`Proxy Found: '${prefix}' => '${proxyTarget}'`);
                break;
            }
        }
        if (!proxyTarget) {
            logger.warn('Proxy NOT FOUND, skipped');
            return next();
        }
        logger.log(`Request '${url}' will be proxied to '${proxyTarget + ctx.url}'`);
        return koaHttpProxy(proxyTarget, {
            // do not parse body, do not limit body size
            parseReqBody: false,
            /**
             * callback before proxy request
             * @param {Object} proxyReqOpts - proxy request options
             * @param {ctx} ctx - koa ctx
             * @return {Promise.<*>} *
             */
            async proxyReqOptDecorator(proxyReqOpts, context) {
                // remove header `referer` and `origin` to avoid CSRF interception problem by backend when development
                if (IS_DEV) {
                    delete proxyReqOpts.headers.referer;
                    delete proxyReqOpts.headers.origin;
                }
                // record the timestamp for proxy duration
                context.state.proxyStartTimestamp = Date.now();
                return proxyReqOpts;
            },
            /**
             * callback after proxy response
             * @param {Response} proxyRes - proxy response
             * @param {Object} proxyResData - response data
             * @param {ctx} ctx - koa ctx
             * @return {Promise.<*>} *
             */
            async userResDecorator(proxyRes, proxyResData, context) {
                logger.info('ProxyRes headers:', proxyRes.headers);
                const location = `${proxyTarget}${context.url}`;
                // tslint:disable-next-line
                logger.log(`Proxy request '${location}' completed(${proxyRes.statusCode}), costing ${Date.now() - context.state.proxyStartTimestamp}ms.`);
                return proxyResData;
            },
        })(ctx, next);
    };
};
//# sourceMappingURL=proxy.js.map